!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    #################### 
     MODULE MODI_TURB_VER_DYN_FLUX 
!    ####################
!
INTERFACE 
!
      SUBROUTINE TURB_VER_DYN_FLUX(KKA,KKU,KKL,                     &
                      OTURB_FLX,KRR,                                &
                      HTURBDIM,PIMPL,PEXPL,                         &
                      PTSTEP,                                       &
                      TPFILE,                                       &
                      PDXX,PDYY,PDZZ,PDZX,PDZY,PDIRCOSZW,           &
                      PCOSSLOPE,PSINSLOPE,                          &
                      PRHODJ,                                       &
                      PCDUEFF,PTAU11M,PTAU12M,PTAU33M,              &
                      PTHLM,PRM,PSVM,PUM,PVM,PWM,PUSLOPEM,PVSLOPEM, &
                      PTKEM,PLM,PWU,PWV,                            &
                      PRUS,PRVS,PRWS,                               &
                      PDP                                           )
!
USE MODD_IO, ONLY: TFILEDATA
!
INTEGER,                INTENT(IN)   :: KKA           !near ground array index  
INTEGER,                INTENT(IN)   :: KKU           !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL           !vert. levels type 1=MNH -1=AR 
LOGICAL,                INTENT(IN)   ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
INTEGER,                INTENT(IN)   ::  KRR          ! number of moist var.
CHARACTER(len=4),       INTENT(IN)   ::  HTURBDIM     ! dimensionality of the
                                                      ! turbulence scheme
REAL,                   INTENT(IN)   ::  PIMPL, PEXPL ! Coef. for temporal disc.
REAL,                   INTENT(IN)   ::  PTSTEP       ! Double Time Step
TYPE(TFILEDATA),        INTENT(INOUT)::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDXX, PDYY, PDZZ, PDZX, PDZY 
                                                      ! Metric coefficients
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PDIRCOSZW    ! Director Cosinus of the
                                                      ! normal to the ground surface
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCOSSLOPE    ! cosinus of the angle 
                                      ! between i and the slope vector
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSINSLOPE    ! sinus of the angle 
                                      ! between i and the slope vector
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRHODJ       ! dry density * grid volum
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCDUEFF     ! Cd * || u || at time t
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU11M      ! <uu> in the axes linked 
       ! to the maximum slope direction and the surface normal and the binormal 
       ! at time t - dt
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU12M      ! <uv> in the same axes
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU33M      ! <ww> in the same axes
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PUM,PVM,PWM,PTHLM
  ! Wind at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PRM
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PSVM
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PUSLOPEM     ! wind component along the 
                                     ! maximum slope direction
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PVSLOPEM     ! wind component along the 
                                     ! direction normal to the maximum slope one
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTKEM        ! TKE at time t
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLM          ! Turb. mixing length   
REAL, DIMENSION(:,:,:), INTENT(OUT)  ::  PWU          ! momentum flux u'w'
REAL, DIMENSION(:,:,:), INTENT(OUT)  ::  PWV          ! momentum flux v'w'
!
REAL, DIMENSION(:,:,:), INTENT(INOUT)   ::  PRUS, PRVS, PRWS
                            ! cumulated sources for the prognostic variables
!
REAL, DIMENSION(:,:,:), INTENT(OUT)  ::  PDP          ! Dynamic TKE production term
!
END SUBROUTINE TURB_VER_DYN_FLUX
!
END INTERFACE
!
END MODULE MODI_TURB_VER_DYN_FLUX
!
!     ###############################################################
      SUBROUTINE TURB_VER_DYN_FLUX(KKA,KKU,KKL,                     &
                      OTURB_FLX,KRR,                                &
                      HTURBDIM,PIMPL,PEXPL,                         &
                      PTSTEP,                                       &
                      TPFILE,                                       &
                      PDXX,PDYY,PDZZ,PDZX,PDZY,PDIRCOSZW,           &
                      PCOSSLOPE,PSINSLOPE,                          &
                      PRHODJ,                                       &
                      PCDUEFF,PTAU11M,PTAU12M,PTAU33M,              &
                      PTHLM,PRM,PSVM,PUM,PVM,PWM,PUSLOPEM,PVSLOPEM, &
                      PTKEM,PLM,PWU,PWV,                            &
                      PRUS,PRVS,PRWS,                               &
                      PDP                                           )
!     ###############################################################
!
!
!!****  *TURB_VER_DYN_FLUX* -compute the source terms due to the vertical turbulent
!!       fluxes.
!!
!!    PURPOSE
!!    -------
!       The purpose of this routine is to compute the vertical turbulent
!     fluxes of the evolutive variables and give back the source 
!     terms to the main program.	In the case of large horizontal meshes,
!     the divergence of these vertical turbulent fluxes represent the whole
!     effect of the turbulence but when the three-dimensionnal version of
!     the turbulence scheme is activated (CTURBDIM="3DIM"), these divergences
!     are completed in the next routine TURB_HOR. 
!		  An arbitrary degree of implicitness has been implemented for the 
!     temporal treatment of these diffusion terms.
!       The vertical boundary conditions are as follows:
!           *  at the bottom, the surface fluxes are prescribed at the same
!              as the other turbulent fluxes
!           *  at the top, the turbulent fluxes are set to 0.
!       It should be noted that the condensation has been implicitely included
!     in this turbulence scheme by using conservative variables and computing
!     the subgrid variance of a statistical variable s indicating the presence 
!     or not of condensation in a given mesh. 
!
!!**  METHOD
!!    ------
!!      1D type calculations are made;
!!      The vertical turbulent fluxes are computed in an off-centered
!!      implicit scheme (a Crank-Nicholson type with coefficients different
!!      than 0.5), which allows to vary the degree of implicitness of the
!!      formulation.
!!      	 The different prognostic variables are treated one by one. 
!!      The contributions of each turbulent fluxes are cumulated into the 
!!      tendency  PRvarS, and into the dynamic and thermal production of 
!!      TKE if necessary.
!!        
!!			 In section 2 and 3, the thermodynamical fields are considered.
!!      Only the turbulent fluxes of the conservative variables
!!      (Thetal and Rnp stored in PRx(:,:,:,1))  are computed. 
!!       Note that the turbulent fluxes at the vertical 
!!      boundaries are given either by the soil scheme for the surface one
!!      ( at the same instant as the others fluxes) and equal to 0 at the 
!!      top of the model. The thermal production is computed by vertically 
!!      averaging the turbulent flux and multiply this flux at the mass point by
!!      a function ETHETA or EMOIST, which preform the transformation from the
!!      conservative variables to the virtual potential temperature. 
!!     
!! 	    In section 4, the variance of the statistical variable
!!      s indicating presence or not of condensation, is determined in function 
!!      of the turbulent moments of the conservative variables and its
!!      squarred root is stored in PSIGS. This information will be completed in 
!!      the horizontal turbulence if the turbulence dimensionality is not 
!!      equal to "1DIM".
!!
!!			 In section 5, the x component of the stress tensor is computed.
!!      The surface flux <u'w'> is computed from the value of the surface
!!      fluxes computed in axes linked to the orography ( i", j" , k"):
!!        i" is parallel to the surface and in the direction of the maximum
!!           slope
!!        j" is also parallel to the surface and in the normal direction of
!!           the maximum slope
!!        k" is the normal to the surface
!!      In order to prevent numerical instability, the implicit scheme has 
!!      been extended to the surface flux regarding to its dependence in 
!!      function of U. The dependence in function of the other components 
!!      introduced by the different rotations is only explicit.
!!      The turbulent fluxes are used to compute the dynamic production of 
!!      TKE. For the last TKE level ( located at PDZZ(:,:,IKB)/2 from the
!!      ground), an harmonic extrapolation from the dynamic production at 
!!      PDZZ(:,:,IKB) is used to avoid an evaluation of the gradient of U
!!      in the surface layer.
!!
!!         In section 6, the same steps are repeated but for the y direction
!!		  and in section 7, a diagnostic computation of the W variance is 
!!      performed.
!!
!!         In section 8, the turbulent fluxes for the scalar variables are 
!!      computed by the same way as the conservative thermodynamical variables
!!
!!            
!!    EXTERNAL
!!    --------
!!      GX_U_M, GY_V_M, GZ_W_M :  cartesian gradient operators 
!!      GX_U_UW,GY_V_VW	         (X,Y,Z) represent the direction of the gradient
!!                               _(M,U,...)_ represent the localization of the 
!!                               field to be derivated
!!                               _(M,UW,...) represent the localization of the 
!!                               field	derivated
!!                               
!!
!!      MXM,MXF,MYM,MYF,MZM,MZF
!!                             :  Shuman functions (mean operators)     
!!      DXF,DYF,DZF,DZM
!!                             :  Shuman functions (difference operators)     
!!                               
!!      SUBROUTINE TRIDIAG     : to compute the split implicit evolution
!!                               of a variable located at a mass point
!!
!!      SUBROUTINE TRIDIAG_WIND: to compute the split implicit evolution
!!                               of a variable located at a wind point
!!
!!      FUNCTIONs ETHETA and EMOIST  :  
!!            allows to compute:
!!            - the coefficients for the turbulent correlation between
!!            any variable and the virtual potential temperature, of its 
!!            correlations with the conservative potential temperature and 
!!            the humidity conservative variable:
!!            -------              -------              -------
!!            A' Thv'  =  ETHETA   A' Thl'  +  EMOIST   A' Rnp'  
!!
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      Module MODD_CST : contains physical constants
!!
!!           XG         : gravity constant
!!
!!      Module MODD_CTURB: contains the set of constants for
!!                        the turbulence scheme
!!
!!           XCMFS,XCMFB : cts for the momentum flux
!!           XCSHF       : ct for the sensible heat flux
!!           XCHF        : ct for the moisture flux
!!           XCTV,XCHV   : cts for the T and moisture variances
!!
!!      Module MODD_PARAMETERS
!!
!!           JPVEXT_TURB     : number of vertical external points
!!           JPHEXT     : number of horizontal external points
!!
!!
!!    REFERENCE
!!    ---------
!!      Book 1 of documentation (Chapter: Turbulence)
!!
!!    AUTHOR
!!    ------
!!      Joan Cuxart             * INM and Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original       August   19, 1994
!!      Modifications: February 14, 1995 (J.Cuxart and J.Stein) 
!!                                  Doctorization and Optimization
!!      Modifications: March 21, 1995 (J.M. Carriere) 
!!                                  Introduction of cloud water
!!      Modifications: June  14, 1995 (J.Cuxart and J. Stein) 
!!                                 Phi3 and Psi3 at w-point + bug in the all
!!                                 or nothing condens. 
!!      Modifications: Sept  15, 1995 (J.Cuxart and J. Stein) 
!!                                 Change the DP computation at the ground
!!      Modifications: October 10, 1995 (J.Cuxart and J. Stein) 
!!                                 Psi for scal var and LES tools
!!      Modifications: November 10, 1995 (J. Stein)
!!                                 change the surface	relations 
!!      Modifications: February 20, 1995 (J. Stein) optimization
!!      Modifications: May 21, 1996 (J. Stein) 
!!                                  bug in the vertical flux of the V wind 
!!                                  component for explicit computation
!!      Modifications: May 21, 1996 (N. wood) 
!!                                  modify the computation of the vertical
!!                                   part or the surface tangential flux
!!      Modifications: May 21, 1996 (P. Jabouille)
!!                                  same modification in the Y direction
!!      
!!      Modifications: Sept 17, 1996 (J. Stein) change the moist case by using
!!                                  Pi instead of Piref + use Atheta and Amoist
!!
!!      Modifications: Nov  24, 1997 (V. Masson) removes the DO loops 
!!      Modifications: Mar  31, 1998 (V. Masson) splits the routine TURB_VER_DYN_FLUX 
!!      Modifications: Oct  18, 2000 (J. Stein)  Bug in some computations for IKB level
!!      Modifications: Oct  18, 2000 (V. Masson) LES computations + LFLAT switch
!!                     Nov  06, 2002 (V. Masson) LES budgets
!!                     October 2009 (G. Tanguy) add ILENCH=LEN(YCOMMENT) after
!!                                              change of YCOMMENT
!!      2012-02 Y. Seity,  add possibility to run with reversed vertical levels
!!      J.Escobar : 15/09/2015 : WENO5 & JPHEXT <> 1 
!!      04/2016 (M.Moge) Use openACC directives to port the TURB part of Meso-NH on GPU
!!  Philippe Wautelet: 05/2016-04/2018: new data structures and calls for I/O
!!      Q. Rodier      17/01/2019 : cleaning : remove cyclic conditions on DP and ZA
!! JL Redelsperger 03/2021 : Add Ocean  & O-A Autocoupling LES Cases
!!--------------------------------------------------------------------------
!       
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CONF
USE MODD_CST
USE MODD_CTURB
USE MODD_DYN_n,          ONLY: LOCEAN
use modd_field,          only: tfielddata, TYPEREAL
USE MODD_IO,             ONLY: TFILEDATA
USE MODD_LES
USE MODD_NSV
USE MODD_OCEANH
USE MODD_PARAMETERS
USE MODD_REF, ONLY : LCOUPLES
USE MODD_TURB_n
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
USE MODI_GRADIENT_M
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
USE MODI_LES_MEAN_SUBGRID
USE MODI_SECOND_MNH
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN !Kept because some parts are not yet ported with OpenACC
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_TRIDIAG
USE MODI_TRIDIAG_WIND
!
USE MODE_IO_FIELD_WRITE, only: IO_Field_write
USE MODE_ll
use mode_mppdb
#ifdef MNH_OPENACC
USE MODE_MSG
#endif
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif

IMPLICIT NONE
!
!*      0.1  declarations of arguments
!
!
!
INTEGER,                INTENT(IN)   :: KKA           !near ground array index  
INTEGER,                INTENT(IN)   :: KKU           !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL           !vert. levels type 1=MNH -1=ARO
LOGICAL,                INTENT(IN)   ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
INTEGER,                INTENT(IN)   ::  KRR          ! number of moist var.
CHARACTER(len=4),       INTENT(IN)   ::  HTURBDIM     ! dimensionality of the
                                                      ! turbulence scheme
REAL,                   INTENT(IN)   ::  PIMPL, PEXPL ! Coef. for temporal disc.
REAL,                   INTENT(IN)   ::  PTSTEP       ! Double Time Step
TYPE(TFILEDATA),        INTENT(INOUT)::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDXX, PDYY, PDZZ, PDZX, PDZY 
                                                      ! Metric coefficients
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PDIRCOSZW    ! Director Cosinus of the
                                                      ! normal to the ground surface
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCOSSLOPE    ! cosinus of the angle 
                                      ! between i and the slope vector
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSINSLOPE    ! sinus of the angle 
                                      ! between i and the slope vector
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PRHODJ       ! dry density * grid volum
!
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCDUEFF     ! Cd * || u || at time t
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU11M      ! <uu> in the axes linked 
       ! to the maximum slope direction and the surface normal and the binormal 
       ! at time t - dt
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU12M      ! <uv> in the same axes
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PTAU33M      ! <ww> in the same axes
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PUM,PVM,PWM, PTHLM
  ! Wind at t-Delta t
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PRM
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::  PSVM
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PUSLOPEM     ! wind component along the 
                                     ! maximum slope direction
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PVSLOPEM     ! wind component along the 
                                     ! direction normal to the maximum slope one
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PTKEM        ! TKE at time t
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PLM          ! Turb. mixing length   
REAL, DIMENSION(:,:,:), INTENT(OUT)  ::  PWU          ! momentum flux u'w'
REAL, DIMENSION(:,:,:), INTENT(OUT)  ::  PWV          ! momentum flux v'w'
!
REAL, DIMENSION(:,:,:), INTENT(INOUT)   ::  PRUS, PRVS, PRWS
                            ! cumulated sources for the prognostic variables
!
REAL, DIMENSION(:,:,:), INTENT(OUT)  ::  PDP          ! Dynamic TKE production term
!
!*       0.2  declaration of local variables
!
!
REAL, DIMENSION(:,:), pointer , contiguous  :: ZDIRSINZW ! sinus of the angle
                   ! between the normal and the vertical at the surface
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZCOEFS    ! coeff. for the
                   ! implicit scheme for the wind at the surface
REAL, DIMENSION(:,:,:), pointer , contiguous ::  &
       ZA, &       ! under diagonal elements of the tri-diagonal matrix involved
                   ! in the temporal implicit scheme (also used to store coefficient
                   ! J in Section 5)
       ZRES, &     ! guess of the treated variable at t+ deltat when the turbu-
                   ! lence is the only source of evolution added to the ones
                   ! considered in ZSOURCE  
       ZFLXZ,  &   ! vertical flux of the treated variable
       ZSOURCE,  & ! source of evolution for the treated variable
       ZKEFF       ! effectif diffusion coeff = LT * SQRT( TKE )
INTEGER             :: IIB,IIE, &   ! I index values for the Beginning and End
                       IJB,IJE, &   ! mass points of the domain in the 3 direct.
                       IKB,IKE      !
INTEGER             :: IKT          ! array size in k direction
INTEGER             :: IKTB,IKTE    ! start, end of k loops in physical domain
INTEGER             :: JSV          ! scalar loop counter
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZCOEFFLXU, &
                                    ZCOEFFLXV, ZUSLOPEM, ZVSLOPEM
                                    ! coefficients for the surface flux
                                    ! evaluation and copy of PUSLOPEM and
                                    ! PVSLOPEM in local 3D arrays
INTEGER             :: IIU,IJU      ! size of array in x,y,z directions
!
REAL :: ZTIME1, ZTIME2
!
#ifdef MNH_OPENACC
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP1_DEVICE,ZTMP2_DEVICE,ZTMP3_DEVICE,ZTMP4_DEVICE
#endif
TYPE(TFIELDDATA) :: TZFIELD
!
INTEGER  :: JIU,JJU,JKU
INTEGER  :: JI,JJ,JK
LOGICAL :: GOCEAN !Intermediate variable used to work around a Cray compiler bug (CCE 13.0.0)
!----------------------------------------------------------------------------

!$acc data present( PDXX, PDYY, PDZZ, PDZX, PDZY, PDIRCOSZW,             &
!$acc &             PCOSSLOPE, PSINSLOPE,                                &
!$acc &             PRHODJ,                                              &
!$acc &             PCDUEFF, PTAU11M, PTAU12M, PTAU33M,                  &
!$acc &             PTHLM, PRM, PSVM, PUM, PVM, PWM, PUSLOPEM, PVSLOPEM, &
!$acc &             PTKEM, PLM, PWU, PWV,                                &
!$acc &             PRUS, PRVS, PRWS,                                    &
!$acc &             PDP                                                  )

if ( mppdb_initialized ) then
  !Check all in arrays
  call Mppdb_check( PDXX,      "Turb_ver_dyn_flux beg:PDXX"      )
  call Mppdb_check( PDYY,      "Turb_ver_dyn_flux beg:PDYY"      )
  call Mppdb_check( PDZZ,      "Turb_ver_dyn_flux beg:PDZZ"      )
  call Mppdb_check( PDZX,      "Turb_ver_dyn_flux beg:PDZX"      )
  call Mppdb_check( PDZY,      "Turb_ver_dyn_flux beg:PDZY"      )
  call Mppdb_check( PDIRCOSZW, "Turb_ver_dyn_flux beg:PDIRCOSZW" )
  call Mppdb_check( PCOSSLOPE, "Turb_ver_dyn_flux beg:PCOSSLOPE" )
  call Mppdb_check( PSINSLOPE, "Turb_ver_dyn_flux beg:PSINSLOPE" )
  call Mppdb_check( PRHODJ,    "Turb_ver_dyn_flux beg:PRHODJ"    )
  call Mppdb_check( PCDUEFF,   "Turb_ver_dyn_flux beg:PCDUEFF"   )
  call Mppdb_check( PTAU11M,   "Turb_ver_dyn_flux beg:PTAU11M"   )
  call Mppdb_check( PTAU12M,   "Turb_ver_dyn_flux beg:PTAU12M"   )
  call Mppdb_check( PTAU33M,   "Turb_ver_dyn_flux beg:PTAU33M"   )
  call Mppdb_check( PUM,       "Turb_ver_dyn_flux beg:PUM"       )
  call Mppdb_check( PVM,       "Turb_ver_dyn_flux beg:PVM"       )
  call Mppdb_check( PWM,       "Turb_ver_dyn_flux beg:PWM"       )
  call Mppdb_check( PTHLM,     "Turb_ver_dyn_flux beg:PTHLM"     )
  call Mppdb_check( PRM,       "Turb_ver_dyn_flux beg:PRM"       )
  call Mppdb_check( PSVM,      "Turb_ver_dyn_flux beg:PSVM"      )
  call Mppdb_check( PUSLOPEM,  "Turb_ver_dyn_flux beg:PUSLOPEM"  )
  call Mppdb_check( PVSLOPEM,  "Turb_ver_dyn_flux beg:PVSLOPEM"  )
  call Mppdb_check( PTKEM,     "Turb_ver_dyn_flux beg:PTKEM"     )
  call Mppdb_check( PLM,       "Turb_ver_dyn_flux beg:PLM"       )
  !Check all inout arrays
  call Mppdb_check( prus, "Turb_ver_dyn_flux beg:prus" )
  call Mppdb_check( prvs, "Turb_ver_dyn_flux beg:prvs" )
  call Mppdb_check( prws, "Turb_ver_dyn_flux beg:prws" )
end if

JIU =  size(pum, 1 )
JJU =  size(pum, 2 )
JKU =  size(pum, 3 )

#ifndef MNH_OPENACC
allocate( zdirsinzw(JIU,JJU ) )

allocate( zcoefs   (JIU,JJU, 1 ) )

allocate( za       (JIU,JJU,JKU ) )
allocate( zres     (JIU,JJU,JKU ) )
allocate( zflxz    (JIU,JJU,JKU ) )
allocate( zsource  (JIU,JJU,JKU ) )
allocate( zkeff    (JIU,JJU,JKU ) )

allocate( zcoefflxu(JIU,JJU, 1 ) )
allocate( zcoefflxv(JIU,JJU, 1 ) )
allocate( zuslopem (JIU,JJU, 1 ) )
allocate( zvslopem (JIU,JJU, 1 ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'TURB_VER_DYN_FLUX' )

CALL MNH_MEM_GET( zdirsinzw, JIU, JJU )

CALL MNH_MEM_GET( zcoefs, 1, JIU, 1, JJU, 1, 1 )

CALL MNH_MEM_GET( za,      JIU, JJU, JKU )
CALL MNH_MEM_GET( zres,    JIU, JJU, JKU )
CALL MNH_MEM_GET( zflxz,   JIU, JJU, JKU )
CALL MNH_MEM_GET( zsource, JIU, JJU, JKU )
CALL MNH_MEM_GET( zkeff,   JIU, JJU, JKU )

CALL MNH_MEM_GET( zcoefflxu, 1, JIU, 1, JJU, 1 , 1 )
CALL MNH_MEM_GET( zcoefflxv, 1, JIU, 1, JJU, 1 , 1 )
CALL MNH_MEM_GET( zuslopem,  1, JIU, 1, JJU, 1 , 1 )
CALL MNH_MEM_GET( zvslopem,  1, JIU, 1, JJU, 1 , 1 )

CALL MNH_MEM_GET( ztmp1_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp2_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp3_device, JIU, JJU, JKU )
CALL MNH_MEM_GET( ztmp4_device, JIU, JJU, JKU )
#endif

!$acc data present( zdirsinzw, zcoefs, za, zres, zflxz, zsource, zkeff, zcoefflxu, zcoefflxv, zuslopem, zvslopem, &
!$acc &            ztmp1_device, ztmp2_device, ztmp3_device, ztmp4_device )

                                    !
!*       1.   PRELIMINARIES
!             -------------
!
IIU=SIZE(PUM,1)
IJU=SIZE(PUM,2)
CALL GET_INDICE_ll (IIB,IJB,IIE,IJE)
IKB=KKA+JPVEXT_TURB*KKL
IKE=KKU-JPVEXT_TURB*KKL
IKT=SIZE(PUM,3)          
IKTB=1+JPVEXT_TURB              
IKTE=IKT-JPVEXT_TURB

GOCEAN = LOCEAN
!
!$acc kernels  present_cr(ZA,PDP,ZSOURCE)
ZA(:,:,:)  = XUNDEF
PDP(:,:,:) = XUNDEF
ZSOURCE(:,:,:) = 0.
!
#ifndef MNH_BITREP
ZDIRSINZW(:,:) = SQRT(1.-PDIRCOSZW(:,:)**2)
#else
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
   ZDIRSINZW(JI,JJ) = SQRT(1.-BR_P2(PDIRCOSZW(JI,JJ)))
END DO
#endif
!$acc end kernels

!  compute the coefficients for the uncentred gradient computation near the 
!  ground
!
#ifndef MNH_OPENACC
ZKEFF(:,:,:) = MZM( PLM(:,:,:) * SQRT(PTKEM(:,:,:)) )
#else
!PASCAL
!$acc kernels present_cr(ZTMP1_DEVICE)
ZTMP1_DEVICE(:,:,:) = PLM(:,:,:) * SQRT(PTKEM(:,:,:))
!$acc end kernels
CALL MZM_DEVICE(ZTMP1_DEVICE,ZKEFF)
#endif
!
!$acc kernels present_cr(ZUSLOPEM,ZVSLOPEM)
ZUSLOPEM(:,:,1)=PUSLOPEM(:,:)
ZVSLOPEM(:,:,1)=PVSLOPEM(:,:)
!$acc end kernels
!
!----------------------------------------------------------------------------
!
!
!*       5.   SOURCES OF U,W WIND COMPONENTS AND PARTIAL DYNAMIC PRODUCTION 
!             -------------------------------------------------------------
!
!*       5.1  Source of U wind component
!
! Preparation of the arguments for TRIDIAG_WIND 
!
#ifndef MNH_OPENACC
ZA(:,:,:)    = -PTSTEP * XCMFS *                              &
              MXM( ZKEFF ) * MXM(MZM( PRHODJ )) / &
#ifndef MNH_BITREP
              MXM( PDZZ )**2
#else
              BR_P2(MXM( PDZZ ))
#endif
#else
CALL MXM_DEVICE( ZKEFF, ZTMP1_DEVICE )
CALL MZM_DEVICE(PRHODJ, ZTMP2_DEVICE )
CALL MXM_DEVICE(ZTMP2_DEVICE,ZTMP3_DEVICE)
CALL MXM_DEVICE( PDZZ, ZTMP4_DEVICE )
!$acc kernels present_cr(ZA)
#ifndef MNH_BITREP
ZA(:,:,:) = -PTSTEP * XCMFS * ZTMP1_DEVICE(:,:,:) * ZTMP3_DEVICE(:,:,:) / ZTMP4_DEVICE(:,:,:)**2
#else
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZA(JI,JJ,JK) = -PTSTEP * XCMFS * ZTMP1_DEVICE(JI,JJ,JK) * ZTMP3_DEVICE(JI,JJ,JK) / BR_P2(ZTMP4_DEVICE(JI,JJ,JK))
END DO
#endif
!$acc end kernels
#endif
!
!
! Compute the source of U wind component 
!
! compute the coefficient between the vertical flux and the 2 components of the 
! wind following the slope
!$acc kernels present_cr(ZCOEFFLXU)
#ifndef MNH_BITREP
ZCOEFFLXU(:,:,1) = PCDUEFF(:,:) * (PDIRCOSZW(:,:)**2 - ZDIRSINZW(:,:)**2) &
     * PCOSSLOPE(:,:)
#else
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )     
   ZCOEFFLXU(JI,JJ,1) = PCDUEFF(JI,JJ) * (BR_P2(PDIRCOSZW(JI,JJ)) - BR_P2(ZDIRSINZW(JI,JJ))) &
        * PCOSSLOPE(JI,JJ)
END DO
#endif
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
   ZCOEFFLXV(JI,JJ,1) = PCDUEFF(JI,JJ) * PDIRCOSZW(JI,JJ) * PSINSLOPE(JI,JJ)
   
   ! prepare the implicit scheme coefficients for the surface flux
   ZCOEFS(JI,JJ,1)=  ZCOEFFLXU(JI,JJ,1) * PCOSSLOPE(JI,JJ) * PDIRCOSZW(JI,JJ)  &
        +ZCOEFFLXV(JI,JJ,1) * PSINSLOPE(JI,JJ)
END DO
!
! average this flux to be located at the U,W vorticity point
#ifndef MNH_OPENACC
ZCOEFS(:,:,1:1)=MXM(ZCOEFS(:,:,1:1) / PDZZ(:,:,IKB:IKB) )
#else
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
   ZTMP1_DEVICE(JI,JJ,1) = ZCOEFS(JI,JJ,1) / PDZZ(JI,JJ,IKB)
END DO
!$acc end kernels
CALL MXM_DEVICE(ZTMP1_DEVICE(:,:,1:1),ZCOEFS(:,:,1:1))
#endif
!
!
! ZSOURCE= FLUX /DZ
IF (GOCEAN) THEN  ! OCEAN MODEL ONLY
#ifdef MNH_OPENACC
  call Print_msg( NVERB_FATAL, 'GEN', 'TURB_VER_DYN_FLUX', 'OpenACC: LOCEAN=T not yet implemented' )
#endif
  ! Sfx flux assumed to be in SI & at vorticity point
  IF (LCOUPLES) THEN  
    ZSOURCE(:,:,IKE:IKE) = XSSUFL_C(:,:,1:1)/PDZZ(:,:,IKE:IKE) &
         *0.5 * ( 1. + MXM(PRHODJ(:,:,KKU:KKU)) / MXM(PRHODJ(:,:,IKE:IKE))) 
  ELSE
    ZSOURCE(:,:,IKE)     = XSSUFL(:,:)
    ZSOURCE(:,:,IKE:IKE) = ZSOURCE (:,:,IKE:IKE) /PDZZ(:,:,IKE:IKE) &
        *0.5 * ( 1. + MXM(PRHODJ(:,:,KKU:KKU)) / MXM(PRHODJ(:,:,IKE:IKE)) )
  ENDIF
  !No flux at the ocean domain bottom
   ZSOURCE(:,:,IKB)           = 0.
   ZSOURCE(:,:,IKTB+1:IKTE-1) = 0
!
ELSE             !ATMOS MODEL ONLY
  IF (LCOUPLES) THEN 
#ifdef MNH_OPENACC
    call Print_msg( NVERB_FATAL, 'GEN', 'TURB_VER_DYN_FLUX', 'OpenACC: LCOUPLES=T not yet implemented' )
#endif
   ZSOURCE(:,:,IKB:IKB) = XSSUFL_C(:,:,1:1)/PDZZ(:,:,IKB:IKB) &
      * 0.5 * ( 1. + MXM(PRHODJ(:,:,KKA:KKA)) / MXM(PRHODJ(:,:,IKB:IKB)) )
  ELSE               
    ! compute the explicit tangential flux at the W point
!$acc kernels present_cr(ZSOURCE,ZTMP1_DEVICE,ZTMP2_DEVICE)
    ZSOURCE(:,:,IKB)     =                                              &
     PTAU11M(:,:) * PCOSSLOPE(:,:) * PDIRCOSZW(:,:) * ZDIRSINZW(:,:) &
     -PTAU12M(:,:) * PSINSLOPE(:,:) * ZDIRSINZW(:,:)                  &
     -PTAU33M(:,:) * PCOSSLOPE(:,:) * ZDIRSINZW(:,:) * PDIRCOSZW(:,:)  
!
    ! add the vertical part or the surface flux at the U,W vorticity point
!
#ifndef MNH_OPENACC
    ZSOURCE(:,:,IKB:IKB) =                                  &
      (   MXM( ZSOURCE(:,:,IKB:IKB)   / PDZZ(:,:,IKB:IKB) ) &
       +  MXM( ZCOEFFLXU(:,:,1:1) / PDZZ(:,:,IKB:IKB)       &
               *ZUSLOPEM(:,:,1:1)                           &
              -ZCOEFFLXV(:,:,1:1) / PDZZ(:,:,IKB:IKB)       &
               *ZVSLOPEM(:,:,1:1)                      )    &
       -  ZCOEFS(:,:,1:1) * PUM(:,:,IKB:IKB) * PIMPL        &
      ) * 0.5 * ( 1. + MXM(PRHODJ(:,:,KKA:KKA)) / MXM(PRHODJ(:,:,IKB:IKB)) )
#else
    ZTMP1_DEVICE(:,:,IKB) = ZSOURCE(:,:,IKB) / PDZZ(:,:,IKB)
    ZTMP2_DEVICE(:,:,1)   = ZCOEFFLXU(:,:,1) / PDZZ(:,:,IKB) &
                  *ZUSLOPEM(:,:,1)                           &
                 -ZCOEFFLXV(:,:,1) / PDZZ(:,:,IKB)           &
                 *ZVSLOPEM(:,:,1)
!$acc end kernels
    CALL MXM_DEVICE( ZTMP1_DEVICE(:,:,IKB:IKB), ZTMP3_DEVICE(:,:,1:1) )
    CALL MXM_DEVICE( ZTMP2_DEVICE(:,:,1:1), ZTMP4_DEVICE(:,:,1:1) )
    CALL MXM_DEVICE(PRHODJ(:,:,KKA:KKA),ZTMP1_DEVICE(:,:,KKA:KKA))
    CALL MXM_DEVICE(PRHODJ(:,:,IKB:IKB),ZTMP2_DEVICE(:,:,IKB:IKB))
!$acc kernels present_cr(ZSOURCE)
    ZSOURCE(:,:,IKB) = ( ZTMP3_DEVICE(:,:,1) + ZTMP4_DEVICE(:,:,1) - ZCOEFS(:,:,1) * PUM(:,:,IKB) * PIMPL  &
                          ) * 0.5 * ( 1. + ZTMP1_DEVICE(:,:,KKA) / ZTMP2_DEVICE(:,:,IKB) )
!$acc end kernels
#endif
  ENDIF 
!
!$acc kernels present_cr(ZSOURCE)
  ZSOURCE(:,:,IKTB+1:IKTE-1) = 0.
  ZSOURCE(:,:,IKE) = 0.
!$acc end kernels
ENDIF !end ocean or atmosphere cases
!
! Obtention of the split U at t+ deltat 
!
#ifndef MNH_OPENACC
CALL TRIDIAG_WIND(KKA,KKU,KKL,PUM,ZA,ZCOEFS(:,:,1),PTSTEP,PEXPL,PIMPL,   &
                  MXM(PRHODJ),ZSOURCE,ZRES)
#else
CALL MXM_DEVICE(PRHODJ,ZTMP1_DEVICE)
CALL TRIDIAG_WIND(KKA,KKU,KKL,PUM,ZA,ZCOEFS(:,:,1),PTSTEP,PEXPL,PIMPL,   &
                  ZTMP1_DEVICE,ZSOURCE,ZRES)
#endif
! 
!  Compute the equivalent tendency for the U wind component
!
#ifndef MNH_OPENACC
PRUS(:,:,:)=PRUS(:,:,:)+MXM(PRHODJ(:,:,:))*(ZRES(:,:,:)-PUM(:,:,:))/PTSTEP
#else
!$acc kernels present_cr(PRUS)
PRUS(:,:,:)=PRUS(:,:,:)+ZTMP1_DEVICE(:,:,:)*(ZRES(:,:,:)-PUM(:,:,:))/PTSTEP
!$acc end kernels
#endif
!
!
!*       5.2  Partial Dynamic Production
!
! vertical flux of the U wind component
!
#ifndef MNH_OPENACC
ZFLXZ(:,:,:)     = -XCMFS * MXM(ZKEFF) * &
                  DZM (PIMPL*ZRES + PEXPL*PUM) / MXM(PDZZ)
#else
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP2_DEVICE(JI,JJ,JK) = PIMPL*ZRES(JI,JJ,JK) + PEXPL*PUM(JI,JJ,JK)
END DO
!$acc end kernels
CALL MXM_DEVICE(ZKEFF,ZTMP1_DEVICE)
CALL DZM_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
CALL MXM_DEVICE(PDZZ,ZTMP4_DEVICE)
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZFLXZ(JI,JJ,JK)     = -XCMFS * ZTMP1_DEVICE(JI,JJ,JK) * ZTMP3_DEVICE(JI,JJ,JK) / ZTMP4_DEVICE(JI,JJ,JK)
END DO
!$acc end kernels
#endif
!
! surface flux 
#ifndef MNH_OPENACC
ZFLXZ(:,:,IKB:IKB)   =   MXM(PDZZ(:,:,IKB:IKB))  *                &
  ( ZSOURCE(:,:,IKB:IKB)                                          &
   +ZCOEFS(:,:,1:1) * ZRES(:,:,IKB:IKB) * PIMPL                   &                
  ) / 0.5 / ( 1. + MXM(PRHODJ(:,:,KKA:KKA)) / MXM(PRHODJ(:,:,IKB:IKB)) )
#else
CALL MXM_DEVICE(PDZZ(:,:,IKB:IKB),ZTMP1_DEVICE(:,:,IKB:IKB))
CALL MXM_DEVICE(PRHODJ(:,:,KKA:KKA),ZTMP2_DEVICE(:,:,KKA:KKA))
CALL MXM_DEVICE(PRHODJ(:,:,IKB:IKB),ZTMP3_DEVICE(:,:,IKB:IKB))
!$acc kernels present_cr(ZFLXZ)
ZFLXZ(:,:,IKB:IKB)   =   ZTMP1_DEVICE(:,:,IKB:IKB)  *                &
  ( ZSOURCE(:,:,IKB:IKB)                                          &
   +ZCOEFS(:,:,1:1) * ZRES(:,:,IKB:IKB) * PIMPL                   &                
  ) / 0.5 / ( 1. + ZTMP2_DEVICE(:,:,KKA:KKA) / ZTMP3_DEVICE(:,:,IKB:IKB) )
#endif
!
ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB) 
!$acc end kernels

IF (GOCEAN) THEN !ocean model at phys sfc (ocean domain top)
  ZFLXZ(:,:,IKE:IKE)   =   MXM(PDZZ(:,:,IKE:IKE))  *                &
                           ZSOURCE(:,:,IKE:IKE)                     &
                           / 0.5 / ( 1. + MXM(PRHODJ(:,:,KKU:KKU)) / MXM(PRHODJ(:,:,IKE:IKE)) )
  ZFLXZ(:,:,KKU) = ZFLXZ(:,:,IKE) 
END IF
!
IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
!$acc update self(ZFLXZ)
  ! stores the U wind component vertical flux
  TZFIELD%CMNHNAME   = 'UW_VFLX'
  TZFIELD%CSTDNAME   = ''
  TZFIELD%CLONGNAME  = 'UW_VFLX'
  TZFIELD%CUNITS     = 'm2 s-2'
  TZFIELD%CDIR       = 'XY'
  TZFIELD%CCOMMENT   = 'U wind component vertical flux'
  TZFIELD%NGRID      = 4
  TZFIELD%NTYPE      = TYPEREAL
  TZFIELD%NDIMS      = 3
  TZFIELD%LTIMEDEP   = .TRUE.
  CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ)
END IF
!
! first part of total momentum flux
!
!$acc kernels present_cr(PWU)
PWU(:,:,:) = ZFLXZ(:,:,:)
!$acc end kernels
!
! Contribution to the dynamic production of TKE
! compute the dynamic production at the mass point
!
#ifndef MNH_OPENACC
PDP(:,:,:) = - MZF( MXF ( ZFLXZ * GZ_U_UW(PUM,PDZZ) )  )
#else
CALL GZ_U_UW_DEVICE(PUM,PDZZ,ZTMP1_DEVICE)
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU )
   ZTMP2_DEVICE(JI,JJ,JK) = ZFLXZ(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK)
END DO
!$acc end kernels
CALL MXF_DEVICE( ZTMP2_DEVICE,ZTMP3_DEVICE )
CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
!$acc kernels present_cr(PDP)
PDP(:,:,:) = - ZTMP4_DEVICE(:,:,:)
!$acc end kernels
#endif
!
! evaluate the dynamic production at w(IKB+KKL) in PDP(IKB)
#ifndef MNH_OPENACC
PDP(:,:,IKB:IKB) = - MXF (                                                      &
  ZFLXZ(:,:,IKB+KKL:IKB+KKL) * (PUM(:,:,IKB+KKL:IKB+KKL)-PUM(:,:,IKB:IKB))  &
                         / MXM(PDZZ(:,:,IKB+KKL:IKB+KKL))                   &
                         )
#else
CALL MXM_DEVICE(PDZZ(:,:,IKB+KKL:IKB+KKL),ZTMP1_DEVICE(:,:,IKB+KKL:IKB+KKL))
!$acc kernels present_cr(ZTMP2_DEVICE)
ZTMP2_DEVICE(:,:,IKB:IKB) = ZFLXZ(:,:,IKB+KKL:IKB+KKL) * (PUM(:,:,IKB+KKL:IKB+KKL)-PUM(:,:,IKB:IKB))  &
                         / ZTMP1_DEVICE(:,:,IKB+KKL:IKB+KKL)
!$acc end kernels
CALL MXF_DEVICE(ZTMP2_DEVICE(:,:,IKB:IKB), ZTMP3_DEVICE(:,:,IKB:IKB)) 
!$acc kernels present_cr(PDP)
PDP(:,:,IKB:IKB) = - ZTMP3_DEVICE(:,:,IKB:IKB)
!$acc end kernels
#endif
!
IF (GOCEAN) THEN
  ! evaluate the dynamic production at w(IKE-KKL) in PDP(IKE)
  PDP(:,:,IKE:IKE) = - MXF (                                                      &
    ZFLXZ(:,:,IKE-KKL:IKE-KKL) * (PUM(:,:,IKE:IKE)-PUM(:,:,IKE-KKL:IKE-KKL))  &
                           / MXM(PDZZ(:,:,IKE-KKL:IKE-KKL))                   &
                           ) 
END IF
!
! Storage in the LES configuration
! 
IF (LLES_CALL) THEN
  CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
  CALL LES_MEAN_SUBGRID( MZF(MXF(ZFLXZ)), X_LES_SUBGRID_WU )
  CALL LES_MEAN_SUBGRID( MZF(MXF(GZ_U_UW(PUM,PDZZ) &
                          & *ZFLXZ)), X_LES_RES_ddxa_U_SBG_UaU )
  CALL LES_MEAN_SUBGRID( XCMFS * ZKEFF, X_LES_SUBGRID_Km )
#else
!$acc data copy(X_LES_SUBGRID_WU,X_LES_RES_ddxa_U_SBG_UaU,X_LES_SUBGRID_Km)
  !
  CALL MXF_DEVICE(ZFLXZ,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_SUBGRID_WU ) 
  !
  CALL GZ_U_UW_DEVICE(PUM,PDZZ,ZTMP1_DEVICE)
  !$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
  !$acc end kernels
  CALL MXF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
  CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE, X_LES_RES_ddxa_U_SBG_UaU )
  !
  !$acc kernels present_cr(ZTMP1_DEVICE)
  ZTMP1_DEVICE(:,:,:) = XCMFS * ZKEFF(:,:,:)
  !$acc end kernels
  CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_SUBGRID_Km )
  !
!$acc end data
#endif
  CALL SECOND_MNH(ZTIME2)
  XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
END IF
!
!*       5.3  Source of W wind component
!
!
IF(HTURBDIM=='3DIM') THEN
  ! Compute the source for the W wind component
                ! used to compute the W source at the ground
!PASCAL
!$acc kernels present_cr(ZFLXZ)
  ZFLXZ(:,:,KKA) = 2 * ZFLXZ(:,:,IKB) - ZFLXZ(:,:,IKB+KKL) ! extrapolation
!$acc end kernels
 IF (GOCEAN) THEN
!$acc kernels present_cr(ZFLXZ)
   ZFLXZ(:,:,KKU) = 2 * ZFLXZ(:,:,IKE) - ZFLXZ(:,:,IKE-KKL) ! extrapolation
!$acc end kernels
 END IF
  !
#ifndef MNH_OPENACC
  IF (.NOT. LFLAT) THEN
    PRWS(:,:,:)= PRWS                                      &
                -DXF( MZM( MXM(PRHODJ) /PDXX(:,:,:) )  * ZFLXZ(:,:,:) )  &
                +DZM( PRHODJ / MZF(PDZZ ) *                &
                      MXF( MZF( ZFLXZ(:,:,:)*PDZX(:,:,:) ) / PDXX(:,:,:) )      &
                    )
  ELSE
    PRWS(:,:,:)= PRWS(:,:,:) -DXF( MZM( MXM(PRHODJ) /PDXX(:,:,:) )  * ZFLXZ(:,:,:) )
  END IF
#else
  CALL MXM_DEVICE(PRHODJ,ZTMP1_DEVICE)
  !$acc kernels
  !$acc_nv loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) /PDXX(JI,JJ,JK)
  END DO
!$acc end kernels
  CALL MZM_DEVICE(ZTMP2_DEVICE,ZTMP3_DEVICE )
  !$acc kernels
  !$acc_nv loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU) 
     ZTMP2_DEVICE(JI,JJ,JK) = ZTMP3_DEVICE(JI,JJ,JK) * ZFLXZ(JI,JJ,JK)
  END DO
!$acc end kernels
  CALL DXF_DEVICE( ZTMP2_DEVICE,ZTMP1_DEVICE )
  IF (.NOT. LFLAT) THEN
    CALL MZF_DEVICE( PDZZ, ZTMP2_DEVICE )
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP3_DEVICE(JI,JJ,JK) = ZFLXZ(JI,JJ,JK)*PDZX(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP3_DEVICE(JI,JJ,JK) = ZTMP4_DEVICE(JI,JJ,JK) / PDXX(JI,JJ,JK)
    END DO   
    !$acc end kernels
    CALL MXF_DEVICE( ZTMP3_DEVICE,ZTMP4_DEVICE )
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP3_DEVICE(JI,JJ,JK) = PRHODJ(JI,JJ,JK) / ZTMP2_DEVICE(JI,JJ,JK) * ZTMP4_DEVICE(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL DZM_DEVICE( ZTMP3_DEVICE, ZTMP2_DEVICE )
    !$acc kernels present_cr(PRWS)
    PRWS(:,:,:)= PRWS(:,:,:) - ZTMP1_DEVICE(:,:,:) + ZTMP2_DEVICE(:,:,:)
    !$acc end kernels
  ELSE
    !$acc kernels present_cr(PRWS)
    PRWS(:,:,:)= PRWS(:,:,:) -ZTMP1_DEVICE(:,:,:)
    !$acc end kernels
  END IF
#endif
  !
  ! Complete the Dynamical production with the W wind component 
  !
#ifndef MNH_OPENACC
  ZA(:,:,:)=-MZF( MXF ( ZFLXZ * GX_W_UW( PWM,PDXX,PDZZ,PDZX) )  )
#else
  CALL GX_W_UW_DEVICE( PWM,PDXX,PDZZ,PDZX,ZTMP1_DEVICE)
  !$acc kernels
  !$acc_nv loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP2_DEVICE(JI,JJ,JK) = ZFLXZ(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK)
  END DO
!$acc end kernels
  CALL MXF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  !$acc kernels
  !$acc_nv loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZA(JI,JJ,JK)=-ZTMP2_DEVICE(JI,JJ,JK)
  END DO
!$acc end kernels
#endif
  !
  !
  ! evaluate the dynamic production at w(IKB+KKL) in PDP(IKB)
#ifndef MNH_OPENACC
  ZA(:,:,IKB:IKB) = - MXF (                                                  &
   ZFLXZ(:,:,IKB+KKL:IKB+KKL) *                                              &
     ( DXM( PWM(:,:,IKB+KKL:IKB+KKL) )                                       &
      -MXM(  (PWM(:,:,IKB+2*KKL:IKB+2*KKL   )-PWM(:,:,IKB+KKL:IKB+KKL))      &
              /(PDZZ(:,:,IKB+2*KKL:IKB+2*KKL)+PDZZ(:,:,IKB+KKL:IKB+KKL))     &
            +(PWM(:,:,IKB+KKL:IKB+KKL)-PWM(:,:,IKB:IKB  ))                   &
              /(PDZZ(:,:,IKB+KKL:IKB+KKL)+PDZZ(:,:,IKB:IKB  ))               &
          )                                                                  &
        * PDZX(:,:,IKB+KKL:IKB+KKL)                                          &
     ) / (0.5*(PDXX(:,:,IKB+KKL:IKB+KKL)+PDXX(:,:,IKB:IKB)))                 &
                          )
#else
  CALL DXM_DEVICE( PWM(:,:,IKB+KKL:IKB+KKL), ZTMP1_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,1) = (PWM(:,:,IKB+2*KKL   )-PWM(:,:,IKB+KKL)) &
              /(PDZZ(:,:,IKB+2*KKL)+PDZZ(:,:,IKB+KKL))        &
              +(PWM(:,:,IKB+KKL)-PWM(:,:,IKB  ))              &
              /(PDZZ(:,:,IKB+KKL)+PDZZ(:,:,IKB  )) 
!$acc end kernels
  CALL MXM_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP3_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,1) = ZFLXZ(:,:,IKB+KKL) *                          &
     ( ZTMP1_DEVICE(:,:,1) - ZTMP3_DEVICE(:,:,1) * PDZX(:,:,IKB+KKL) )   &
     / (0.5*(PDXX(:,:,IKB+KKL)+PDXX(:,:,IKB)))
!$acc end kernels
  CALL MXF_DEVICE(ZTMP2_DEVICE(:,:,1:1),ZTMP4_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZA)
  ZA(:,:,IKB:IKB) = - ZTMP4_DEVICE(:,:,1:1)
!$acc end kernels
#endif
  !
IF (GOCEAN) THEN
  ! evaluate the dynamic production at w(IKE-KKL) in PDP(IKE)
  ZA(:,:,IKE:IKE) = - MXF (                                                  &
   ZFLXZ(:,:,IKE-KKL:IKE-KKL) *                                              &
     ( DXM( PWM(:,:,IKE-KKL:IKE-KKL) )                                       &
      -MXM(  (PWM(:,:,IKE-2*KKL:IKE-2*KKL   )-PWM(:,:,IKE-KKL:IKE-KKL))      &
              /(PDZZ(:,:,IKE-2*KKL:IKE-2*KKL)+PDZZ(:,:,IKE-KKL:IKE-KKL))     &
            +(PWM(:,:,IKE-KKL:IKE-KKL)-PWM(:,:,IKE:IKE  ))                   &
              /(PDZZ(:,:,IKE-KKL:IKE-KKL)+PDZZ(:,:,IKE:IKE  ))               &
          )                                                                  &
         * PDZX(:,:,IKE-KKL:IKE-KKL)                                          &
     ) / (0.5*(PDXX(:,:,IKE-KKL:IKE-KKL)+PDXX(:,:,IKE:IKE)))                 &
                          )
END IF
  !
!$acc kernels present_cr(PDP)
  PDP(:,:,:)=PDP(:,:,:)+ZA(:,:,:)
!$acc end kernels
  !
  ! Storage in the LES configuration
  ! 
  IF (LLES_CALL) THEN
    CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
    CALL LES_MEAN_SUBGRID( MZF(MXF(GX_W_UW(PWM,PDXX,&
      PDZZ,PDZX)*ZFLXZ)), X_LES_RES_ddxa_W_SBG_UaW )
    CALL LES_MEAN_SUBGRID( MXF(GX_M_U(KKA,KKU,KKL,PTHLM,PDXX,PDZZ,PDZX)&
      * MZF(ZFLXZ)), X_LES_RES_ddxa_Thl_SBG_UaW )
    IF (KRR>=1) THEN
      CALL LES_MEAN_SUBGRID(MXF(GX_U_M(PRM(:,:,:,1),PDXX,PDZZ,PDZX)&
      *MZF(ZFLXZ)),X_LES_RES_ddxa_Rt_SBG_UaW )
    END IF
    DO JSV=1,NSV
      CALL LES_MEAN_SUBGRID( MXF(GX_U_M(PSVM(:,:,:,JSV),PDXX,PDZZ,&
      PDZX)*MZF(ZFLXZ)),X_LES_RES_ddxa_Sv_SBG_UaW(:,:,:,JSV) )
    END DO
#else
!$acc data copy(X_LES_RES_ddxa_W_SBG_UaW,X_LES_RES_ddxa_Thl_SBG_UaW, &
!$acc &         X_LES_RES_ddxa_Rt_SBG_UaW,X_LES_RES_ddxa_Sv_SBG_UaW)
    !
    CALL GX_W_UW_DEVICE( PWM,PDXX,PDZZ,PDZX,ZTMP1_DEVICE)
    !$acc kernels present_cr(ZTMP2_DEVICE)
    ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL MXF_DEVICE(ZTMP2_DEVICE,ZTMP3_DEVICE)
    CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
    CALL LES_MEAN_SUBGRID(ZTMP4_DEVICE, X_LES_RES_ddxa_W_SBG_UaW )
    !
    CALL GX_M_U_DEVICE(KKA,KKU,KKL,PTHLM,PDXX,PDZZ,PDZX,ZTMP1_DEVICE)
    CALL MZF_DEVICE( ZFLXZ, ZTMP2_DEVICE )
    !$acc kernels present_cr(ZTMP3_DEVICE)
    ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
    !$acc end kernels
    CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
    CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE, X_LES_RES_ddxa_Thl_SBG_UaW )
    !
    IF (KRR>=1) THEN
      CALL GX_U_M_DEVICE(PRM(:,:,:,1),PDXX,PDZZ,PDZX,ZTMP1_DEVICE)
      CALL MZF_DEVICE( ZFLXZ, ZTMP2_DEVICE )
      !$acc kernels present_cr(ZTMP3_DEVICE)
      ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
      !$acc end kernels
      CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
      CALL LES_MEAN_SUBGRID(ZTMP4_DEVICE,X_LES_RES_ddxa_Rt_SBG_UaW )
    END IF
    !
    DO JSV=1,NSV
      CALL GX_U_M_DEVICE(PSVM(:,:,:,JSV),PDXX,PDZZ,PDZX,ZTMP1_DEVICE)
      CALL MZF_DEVICE( ZFLXZ, ZTMP2_DEVICE )
      !$acc kernels present_cr(ZTMP3_DEVICE)
      ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
      !$acc end kernels
      CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
      CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE,X_LES_RES_ddxa_Sv_SBG_UaW(:,:,:,JSV) )
    END DO
    !
!$acc end data
#endif
    CALL SECOND_MNH(ZTIME2)
    XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
  END IF
END IF
!
!----------------------------------------------------------------------------
!
!
!*       6.   SOURCES OF V,W WIND COMPONENTS AND COMPLETE 1D DYNAMIC PRODUCTION 
!             -----------------------------------------------------------------
!
!*       6.1  Source of V wind component
!
! Preparation of the arguments for TRIDIAG_WIND
!!
#ifndef MNH_OPENACC
ZA(:,:,:)    = - PTSTEP * XCMFS *                              &
              MYM( ZKEFF ) * MYM(MZM( PRHODJ )) /  &
#ifndef MNH_BITREP
              MYM( PDZZ )**2
#else
              BR_P2(MYM( PDZZ ))
#endif
#else
CALL MYM_DEVICE( ZKEFF, ZTMP1_DEVICE )
CALL MYM_DEVICE( PDZZ, ZTMP2_DEVICE )
CALL MZM_DEVICE(PRHODJ, ZTMP3_DEVICE )
CALL MYM_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
!$acc kernels present_cr(ZA,ZCOEFFLXU)
#ifndef MNH_BITREP
ZA(:,:,:)    = - PTSTEP * XCMFS * ZTMP1_DEVICE(:,:,:) * ZTMP4_DEVICE(:,:,:) / ZTMP2_DEVICE(:,:,:)**2
#else
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZA(JI,JJ,JK)    = - PTSTEP * XCMFS * ZTMP1_DEVICE(JI,JJ,JK) * ZTMP4_DEVICE(JI,JJ,JK) / BR_P2(ZTMP2_DEVICE(JI,JJ,JK))
END DO
#endif
#endif
!
!
!
! Compute the source of V wind component
! compute the coefficient between the vertical flux and the 2 components of the 
! wind following the slope
#ifndef MNH_BITREP
ZCOEFFLXU(:,:,1) = PCDUEFF(:,:) * (PDIRCOSZW(:,:)**2 - ZDIRSINZW(:,:)**2) &
                                   * PSINSLOPE(:,:)
#else
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
ZCOEFFLXU(JI,JJ,1) = PCDUEFF(JI,JJ) * (BR_P2(PDIRCOSZW(JI,JJ)) - BR_P2(ZDIRSINZW(JI,JJ))) &
     * PSINSLOPE(JI,JJ)
END DO
#endif
ZCOEFFLXV(:,:,1) = PCDUEFF(:,:) * PDIRCOSZW(:,:) * PCOSSLOPE(:,:)

! prepare the implicit scheme coefficients for the surface flux
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
ZCOEFS(JI,JJ,1)=  ZCOEFFLXU(JI,JJ,1) * PSINSLOPE(JI,JJ) * PDIRCOSZW(JI,JJ)  &
     +ZCOEFFLXV(JI,JJ,1) * PCOSSLOPE(JI,JJ)
END DO
!
! average this flux to be located at the V,W vorticity point
#ifndef MNH_OPENACC
ZCOEFS(:,:,1:1)=MYM(ZCOEFS(:,:,1:1) / PDZZ(:,:,IKB:IKB) )
#else
!$acc_nv loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU )
   ZTMP1_DEVICE(JI,JJ,1) = ZCOEFS(JI,JJ,1) / PDZZ(JI,JJ,IKB)
END DO
!$acc end kernels
CALL MYM_DEVICE(ZTMP1_DEVICE(:,:,1:1),ZCOEFS(:,:,1:1) )
#endif
!
IF (GOCEAN) THEN ! Ocean case
  IF (LCOUPLES) THEN
    ZSOURCE(:,:,IKE:IKE) =  XSSVFL_C(:,:,1:1)/PDZZ(:,:,IKE:IKE) &
        *0.5 * ( 1. + MYM(PRHODJ(:,:,KKU:KKU)) / MYM(PRHODJ(:,:,IKE:IKE)) ) 
  ELSE 
    ZSOURCE(:,:,IKE) = XSSVFL(:,:)
    ZSOURCE(:,:,IKE:IKE) = ZSOURCE(:,:,IKE:IKE)/PDZZ(:,:,IKE:IKE) &
        *0.5 * ( 1. + MYM(PRHODJ(:,:,KKU:KKU)) / MYM(PRHODJ(:,:,IKE:IKE)) )
  END IF
  !No flux at the ocean domain bottom
  ZSOURCE(:,:,IKB) = 0.
ELSE ! Atmos case
  IF (.NOT.LCOUPLES) THEN !  only atmosp without coupling
  ! compute the explicit tangential flux at the W point
!$acc kernels present_cr(ZSOURCE,ZTMP1_DEVICE)
    ZSOURCE(:,:,IKB)       =                                                  &
      PTAU11M(:,:) * PSINSLOPE(:,:) * PDIRCOSZW(:,:) * ZDIRSINZW(:,:)         &
     +PTAU12M(:,:) * PCOSSLOPE(:,:) * ZDIRSINZW(:,:)                          &
     -PTAU33M(:,:) * PSINSLOPE(:,:) * ZDIRSINZW(:,:) * PDIRCOSZW(:,:) 
!
  ! add the vertical part or the surface flux at the V,W vorticity point
#ifndef MNH_OPENACC
    ZSOURCE(:,:,IKB:IKB) =                                      &
      (   MYM( ZSOURCE(:,:,IKB:IKB)   / PDZZ(:,:,IKB:IKB) )     &
       +  MYM( ZCOEFFLXU(:,:,1:1) / PDZZ(:,:,IKB:IKB)           &
              *ZUSLOPEM(:,:,1:1)                                &
              +ZCOEFFLXV(:,:,1:1) / PDZZ(:,:,IKB:IKB)           &
              *ZVSLOPEM(:,:,1:1)                      )         &
       - ZCOEFS(:,:,1:1) * PVM(:,:,IKB:IKB) * PIMPL             &
      ) * 0.5 * ( 1. + MYM(PRHODJ(:,:,KKA:KKA)) / MYM(PRHODJ(:,:,IKB:IKB)) )
#else
    ZTMP1_DEVICE(:,:,1) = ZSOURCE(:,:,IKB)   / PDZZ(:,:,IKB)
!$acc end kernels
    CALL MYM_DEVICE( ZTMP1_DEVICE(:,:,1:1), ZTMP2_DEVICE(:,:,1:1) )
!$acc kernels present_cr(ZTMP1_DEVICE)
    ZTMP1_DEVICE(:,:,1) = ZCOEFFLXU(:,:,1) / PDZZ(:,:,IKB) &
              *ZUSLOPEM(:,:,1)                          &
              +ZCOEFFLXV(:,:,1) / PDZZ(:,:,IKB)         &
              *ZVSLOPEM(:,:,1)
!$acc end kernels
    CALL MYM_DEVICE(ZTMP1_DEVICE(:,:,1:1),ZTMP3_DEVICE(:,:,1:1))
    CALL MYM_DEVICE(PRHODJ(:,:,KKA:KKA),ZTMP1_DEVICE(:,:,1:1))
    CALL MYM_DEVICE(PRHODJ(:,:,IKB:IKB),ZTMP4_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZSOURCE)
    ZSOURCE(:,:,IKB) = ( ZTMP2_DEVICE(:,:,1) +  ZTMP3_DEVICE(:,:,1) - ZCOEFS(:,:,1) * PVM(:,:,IKB) * PIMPL ) &
      * 0.5 * ( 1. + ZTMP1_DEVICE(:,:,1) / ZTMP4_DEVICE(:,:,1) )
!$acc end kernels
#endif
!
  ELSE   !atmosphere when coupling
    ! input flux assumed to be in SI and at vorticity point
    ZSOURCE(:,:,IKB:IKB) =     -XSSVFL_C(:,:,1:1)/(1.*PDZZ(:,:,IKB:IKB)) &
      * 0.5 * ( 1. + MYM(PRHODJ(:,:,KKA:KKA)) / MYM(PRHODJ(:,:,IKB:IKB)) )
  ENDIF
  !No flux at the atmosphere top
!$acc kernels present_cr(ZSOURCE)
  ZSOURCE(:,:,IKE) = 0.
!$acc end kernels
ENDIF ! End of Ocean or Atmospher Cases
!$acc kernels present_cr(ZSOURCE)
ZSOURCE(:,:,IKTB+1:IKTE-1) = 0.
!$acc end kernels
!
!  Obtention of the split V at t+ deltat 
#ifndef MNH_OPENACC
CALL TRIDIAG_WIND(KKA,KKU,KKL,PVM,ZA,ZCOEFS(:,:,1),PTSTEP,PEXPL,PIMPL,  &
                  MYM(PRHODJ),ZSOURCE,ZRES)
#else
CALL MYM_DEVICE(PRHODJ,ZTMP1_DEVICE)
CALL TRIDIAG_WIND(KKA,KKU,KKL,PVM,ZA,ZCOEFS(:,:,1),PTSTEP,PEXPL,PIMPL,  &
                  ZTMP1_DEVICE,ZSOURCE,ZRES)
#endif
!
! Compute the equivalent tendency for the V wind component
!
#ifndef MNH_OPENACC
PRVS(:,:,:)=PRVS(:,:,:)+MYM(PRHODJ(:,:,:))*(ZRES(:,:,:)-PVM(:,:,:))/PTSTEP
#else
!$acc kernels present_cr(PRVS)
PRVS(:,:,:)=PRVS(:,:,:)+ZTMP1_DEVICE*(ZRES(:,:,:)-PVM(:,:,:))/PTSTEP
!$acc end kernels
#endif
!
!
!*       6.2  Complete 1D dynamic Production
!
!  vertical flux of the V wind component
!
#ifndef MNH_OPENACC
ZFLXZ(:,:,:)   = -XCMFS * MYM(ZKEFF) * &
                DZM( PIMPL*ZRES + PEXPL*PVM ) / MYM(PDZZ)
!
ZFLXZ(:,:,IKB:IKB)   =   MYM(PDZZ(:,:,IKB:IKB))  *                       &
  ( ZSOURCE(:,:,IKB:IKB)                                                 &
   +ZCOEFS(:,:,1:1) * ZRES(:,:,IKB:IKB) * PIMPL                      &      
  ) / 0.5 / ( 1. + MYM(PRHODJ(:,:,KKA:KKA)) / MYM(PRHODJ(:,:,IKB:IKB)) )
#else
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP1_DEVICE(JI,JJ,JK) = PIMPL*ZRES(JI,JJ,JK) + PEXPL*PVM(JI,JJ,JK)
END DO
!$acc end kernels
CALL DZM_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
CALL MYM_DEVICE(PDZZ,ZTMP3_DEVICE)
CALL MYM_DEVICE(ZKEFF,ZTMP1_DEVICE)
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZFLXZ(JI,JJ,JK)   = -XCMFS * ZTMP1_DEVICE(JI,JJ,JK) * ZTMP2_DEVICE(JI,JJ,JK) / ZTMP3_DEVICE(JI,JJ,JK)
END DO
!$acc end kernels
!
CALL MYM_DEVICE(PDZZ(:,:,IKB:IKB),ZTMP1_DEVICE(:,:,1:1))
CALL MYM_DEVICE(PRHODJ(:,:,KKA:KKA),ZTMP2_DEVICE(:,:,1:1))
CALL MYM_DEVICE(PRHODJ(:,:,IKB:IKB),ZTMP3_DEVICE(:,:,1:1))
!$acc kernels present_cr(ZFLXZ)
ZFLXZ(:,:,IKB)   =   ZTMP1_DEVICE(:,:,1)  *                 &
  ( ZSOURCE(:,:,IKB)                                     &
   +ZCOEFS(:,:,1) * ZRES(:,:,IKB) * PIMPL                &      
  ) / 0.5 / ( 1. + ZTMP2_DEVICE(:,:,1) / ZTMP3_DEVICE(:,:,1) )
#endif
!
ZFLXZ(:,:,KKA) = ZFLXZ(:,:,IKB)
!$acc end kernels
!
IF (GOCEAN) THEN
  ZFLXZ(:,:,IKE:IKE)   =   MYM(PDZZ(:,:,IKE:IKE))  *                &
      ZSOURCE(:,:,IKE:IKE)                                          &
      / 0.5 / ( 1. + MYM(PRHODJ(:,:,KKU:KKU)) / MYM(PRHODJ(:,:,IKE:IKE)) )
  ZFLXZ(:,:,KKU) = ZFLXZ(:,:,IKE)
END IF
!
IF ( OTURB_FLX .AND. tpfile%lopened ) THEN
!$acc update self(ZFLXZ)
  ! stores the V wind component vertical flux
  TZFIELD%CMNHNAME   = 'VW_VFLX'
  TZFIELD%CSTDNAME   = ''
  TZFIELD%CLONGNAME  = 'VW_VFLX'
  TZFIELD%CUNITS     = 'm2 s-2'
  TZFIELD%CDIR       = 'XY'
  TZFIELD%CCOMMENT   = 'V wind component vertical flux'
  TZFIELD%NGRID      = 4
  TZFIELD%NTYPE      = TYPEREAL
  TZFIELD%NDIMS      = 3
  TZFIELD%LTIMEDEP   = .TRUE.
  CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ)
END IF
!
! second part of total momentum flux
!
!$acc kernels present_cr(PWV)
PWV(:,:,:) = ZFLXZ(:,:,:)
!$acc end kernels
!
!  Contribution to the dynamic production of TKE
! compute the dynamic production contribution at the mass point
!
#ifndef MNH_OPENACC
ZA(:,:,:) = - MZF( MYF ( ZFLXZ * GZ_V_VW(PVM,PDZZ) ) )
#else
CALL GZ_V_VW_DEVICE(PVM,PDZZ,ZTMP1_DEVICE)
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZTMP2_DEVICE(JI,JJ,JK) = ZFLXZ(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK)
END DO
!$acc end kernels
CALL MYF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP1_DEVICE )
!$acc kernels
!$acc_nv loop independent collapse(3)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
   ZA(JI,JJ,JK) = - ZTMP1_DEVICE(JI,JJ,JK)
END DO   
!$acc end kernels
#endif
!
! evaluate the dynamic production at w(IKB+KKL) in PDP(IKB)
#ifndef MNH_OPENACC
ZA(:,:,IKB:IKB)  =                                                 &
                 - MYF (                                          &
ZFLXZ(:,:,IKB+KKL:IKB+KKL) * (PVM(:,:,IKB+KKL:IKB+KKL)-PVM(:,:,IKB:IKB))  &
                       / MYM(PDZZ(:,:,IKB+KKL:IKB+KKL))               &
                       )
#else
CALL MYM_DEVICE( PDZZ(:,:,IKB+KKL:IKB+KKL), ZTMP1_DEVICE(:,:,1:1) )
!$acc kernels present_cr(ZTMP2_DEVICE)
ZTMP2_DEVICE(:,:,1) = ZFLXZ(:,:,IKB+KKL) * (PVM(:,:,IKB+KKL)-PVM(:,:,IKB)) / ZTMP1_DEVICE(:,:,1)
!$acc end kernels
CALL MYF_DEVICE( ZTMP2_DEVICE(:,:,1:1), ZTMP1_DEVICE(:,:,1:1) )
!$acc kernels present_cr(ZA)
ZA(:,:,IKB)  = - ZTMP1_DEVICE(:,:,1)
!$acc end kernels
#endif
!
IF (GOCEAN) THEN
  ! evaluate the dynamic production at w(IKE-KKL) in PDP(IKE)
  ZA(:,:,IKE:IKE) = - MYF (                                                  &
   ZFLXZ(:,:,IKE-KKL:IKE-KKL) * (PVM(:,:,IKE:IKE)-PVM(:,:,IKE-KKL:IKE-KKL))  &
                          / MYM(PDZZ(:,:,IKE-KKL:IKE-KKL))                   &
                          )
END IF
!
!$acc kernels present_cr(PDP)
PDP(:,:,:)=PDP(:,:,:)+ZA(:,:,:)
!$acc end kernels
!
! Storage in the LES configuration
!
IF (LLES_CALL) THEN
  CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
  CALL LES_MEAN_SUBGRID( MZF(MYF(ZFLXZ)), X_LES_SUBGRID_WV )
  CALL LES_MEAN_SUBGRID( MZF(MYF(GZ_V_VW(PVM,PDZZ)*&
                    & ZFLXZ)), X_LES_RES_ddxa_V_SBG_UaV )
#else
!$acc data copy(X_LES_SUBGRID_WV,X_LES_RES_ddxa_V_SBG_UaV)
  !
  CALL MYF_DEVICE(ZFLXZ,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_SUBGRID_WV ) 
  !
  CALL GZ_V_VW_DEVICE(PVM,PDZZ,ZTMP1_DEVICE)
!$acc kernels present_cr(ZTMP2_DEVICE)
  ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
!$acc end kernels
  CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP3_DEVICE)
  CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE, X_LES_RES_ddxa_V_SBG_UaV )
  !
!$acc end data
#endif
  CALL SECOND_MNH(ZTIME2)
  XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
END IF
!
!
!*       6.3  Source of W wind component 
!
IF(HTURBDIM=='3DIM') THEN
  ! Compute the source for the W wind component
  !PASCAL
!$acc kernels present_cr(ZFLXZ)
  ZFLXZ(:,:,KKA) = 2 * ZFLXZ(:,:,IKB) - ZFLXZ(:,:,IKB+KKL) ! extrapolation 
!$acc end kernels
  IF (GOCEAN) THEN
!$acc kernels present_cr(ZFLXZ)
    ZFLXZ(:,:,KKU) = 2 * ZFLXZ(:,:,IKE) - ZFLXZ(:,:,IKE-KKL) ! extrapolation
!$acc end kernels
  END IF
  !
#ifndef MNH_OPENACC
  IF (.NOT. L2D) THEN 
    IF (.NOT. LFLAT) THEN
      PRWS(:,:,:)= PRWS(:,:,:)                               &
                  -DYF( MZM( MYM(PRHODJ) /PDYY ) * ZFLXZ )   &
                  +DZM( PRHODJ / MZF(PDZZ ) *                &
                        MYF( MZF( ZFLXZ*PDZY ) / PDYY )      &
                      )
    ELSE
      PRWS(:,:,:)= PRWS(:,:,:) -DYF( MZM( MYM(PRHODJ) /PDYY ) * ZFLXZ )
    END IF
  END IF
#else
  IF (.NOT. L2D) THEN 
    CALL MYM_DEVICE(PRHODJ,ZTMP1_DEVICE)
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) /PDYY(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL MZM_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * ZFLXZ(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL DYF_DEVICE( ZTMP2_DEVICE,ZTMP1_DEVICE )
    IF (.NOT. LFLAT) THEN
      CALL MZF_DEVICE( PDZZ, ZTMP2_DEVICE )
      !$acc kernels
      !$acc_nv loop independent collapse(3)
      DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
         ZTMP3_DEVICE(JI,JJ,JK) = ZFLXZ(JI,JJ,JK)*PDZY(JI,JJ,JK)
      END DO
      !$acc end kernels
      CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
      !$acc kernels
      !$acc_nv loop independent collapse(3)
      DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
         ZTMP3_DEVICE(JI,JJ,JK) = ZTMP4_DEVICE(JI,JJ,JK) / PDYY(JI,JJ,JK)
      END DO
      !$acc end kernels
      CALL MYF_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
      !$acc kernels
      !$acc_nv loop independent collapse(3)
      DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
         ZTMP3_DEVICE(JI,JJ,JK) =  PRHODJ(JI,JJ,JK) / ZTMP2_DEVICE(JI,JJ,JK) * ZTMP4_DEVICE(JI,JJ,JK)
      END DO
      !$acc end kernels
      CALL DZM_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
      !$acc kernels present_cr(PRWS)
      PRWS(:,:,:)= PRWS(:,:,:) - ZTMP1_DEVICE(:,:,:) + ZTMP4_DEVICE(:,:,:)
      !$acc end kernels
    ELSE
      !$acc kernels present_cr(PRWS)
      PRWS(:,:,:)= PRWS(:,:,:) - ZTMP1_DEVICE(:,:,:)
      !$acc end kernels
    END IF
  END IF
#endif
  ! 
  ! Complete the Dynamical production with the W wind component 
  IF (.NOT. L2D) THEN
#ifndef MNH_OPENACC
    ZA(:,:,:) = - MZF( MYF ( ZFLXZ(:,:,:) * GY_W_VW( PWM,PDYY,PDZZ,PDZY) )  )
#else
    CALL GY_W_VW_DEVICE( PWM,PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZFLXZ(JI,JJ,JK) * ZTMP1_DEVICE(JI,JJ,JK)
    END DO
    !$acc end kernels
    CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
    CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
    !$acc kernels
    !$acc_nv loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZA(JI,JJ,JK) = - ZTMP2_DEVICE(JI,JJ,JK)
    END DO
    !$acc end kernels
#endif
  !
  ! evaluate the dynamic production at w(IKB+KKL) in PDP(IKB)
#ifndef MNH_OPENACC
    ZA(:,:,IKB:IKB) = - MYF (                                              &
     ZFLXZ(:,:,IKB+KKL:IKB+KKL) *                                          &
       ( DYM( PWM(:,:,IKB+KKL:IKB+KKL) )                                   &
        -MYM(  (PWM(:,:,IKB+2*KKL:IKB+2*KKL)-PWM(:,:,IKB+KKL:IKB+KKL))     &
                /(PDZZ(:,:,IKB+2*KKL:IKB+2*KKL)+PDZZ(:,:,IKB+KKL:IKB+KKL)) &
              +(PWM(:,:,IKB+KKL:IKB+KKL)-PWM(:,:,IKB:IKB  ))               &
                /(PDZZ(:,:,IKB+KKL:IKB+KKL)+PDZZ(:,:,IKB:IKB  ))           &
            )                                                              &
          * PDZY(:,:,IKB+KKL:IKB+KKL)                                      &
       ) / (0.5*(PDYY(:,:,IKB+KKL:IKB+KKL)+PDYY(:,:,IKB:IKB)))             &
                            )
#else
    CALL DYM_DEVICE( PWM(:,:,IKB+KKL:IKB+KKL), ZTMP1_DEVICE(:,:,1:1) )
!$acc kernels present_cr(ZTMP2_DEVICE)
    ZTMP2_DEVICE(:,:,1) = (PWM(:,:,IKB+2*KKL)-PWM(:,:,IKB+KKL)) &
                /(PDZZ(:,:,IKB+2*KKL)+PDZZ(:,:,IKB+KKL))     &
              +(PWM(:,:,IKB+KKL)-PWM(:,:,IKB  ))             &
                /(PDZZ(:,:,IKB+KKL)+PDZZ(:,:,IKB  ))
!$acc end kernels
    CALL MYM_DEVICE( ZTMP2_DEVICE(:,:,1:1), ZTMP3_DEVICE(:,:,1:1) )
!$acc kernels present_cr(ZTMP2_DEVICE)
    ZTMP2_DEVICE(:,:,1) = ZFLXZ(:,:,IKB+KKL) *                        &
       ( ZTMP1_DEVICE(:,:,1) - ZTMP3_DEVICE(:,:,1) * PDZY(:,:,IKB+KKL) ) &
       / (0.5*(PDYY(:,:,IKB+KKL)+PDYY(:,:,IKB)))
!$acc end kernels
    CALL MYF_DEVICE( ZTMP2_DEVICE(:,:,1:1), ZTMP1_DEVICE(:,:,1:1) )
!$acc kernels present_cr(ZA)
    ZA(:,:,IKB:IKB) = - ZTMP1_DEVICE(:,:,1:1)
!$acc end kernels
#endif
  !
    IF (GOCEAN) THEN
     ZA(:,:,IKE:IKE) = - MYF (                                              &
      ZFLXZ(:,:,IKE-KKL:IKE-KKL) *                                          &
        ( DYM( PWM(:,:,IKE-KKL:IKE-KKL) )                                   &
         -MYM(  (PWM(:,:,IKE-2*KKL:IKE-2*KKL)-PWM(:,:,IKE-KKL:IKE-KKL))     &
                 /(PDZZ(:,:,IKE-2*KKL:IKE-2*KKL)+PDZZ(:,:,IKE-KKL:IKE-KKL)) &
               +(PWM(:,:,IKE-KKL:IKE-KKL)-PWM(:,:,IKE:IKE  ))               &
                 /(PDZZ(:,:,IKE-KKL:IKE-KKL)+PDZZ(:,:,IKE:IKE  ))           &
             )                                                              &
           * PDZY(:,:,IKE-KKL:IKE-KKL)                                      &
        ) / (0.5*(PDYY(:,:,IKE-KKL:IKE-KKL)+PDYY(:,:,IKE:IKE)))             &
                            )
    END IF
!    
!$acc kernels present_cr(PDP)
    PDP(:,:,:)=PDP(:,:,:)+ZA(:,:,:)
!$acc end kernels
  !
  END IF
  !
  ! Storage in the LES configuration
  !
  IF (LLES_CALL) THEN
    CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
    CALL LES_MEAN_SUBGRID( MZF(MYF(GY_W_VW(PWM,PDYY,&
    PDZZ,PDZY)*ZFLXZ(:,:,:))), X_LES_RES_ddxa_W_SBG_UaW , .TRUE. )
    CALL LES_MEAN_SUBGRID( MYF(GY_M_V(KKA,KKU,KKL,PTHLM,PDYY,PDZZ,PDZY)&
    *MZF(ZFLXZ)), X_LES_RES_ddxa_Thl_SBG_UaW , .TRUE. )
    IF (KRR>=1) THEN
      CALL LES_MEAN_SUBGRID( MYF(GY_V_M(PRM(:,:,:,1),PDYY,PDZZ,&
      PDZY)*MZF(ZFLXZ)),X_LES_RES_ddxa_Rt_SBG_UaW , .TRUE. )
    END IF
#else
!$acc data copy(X_LES_RES_ddxa_W_SBG_UaW,X_LES_RES_ddxa_Thl_SBG_UaW,X_LES_RES_ddxa_Rt_SBG_UaW)
    !
    CALL GY_W_VW_DEVICE( PWM,PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
    !$acc kernels present_cr(ZTMP2_DEVICE)
    ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLXZ(:,:,:)
    !$acc end kernels
    CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP3_DEVICE)
    CALL MZF_DEVICE( ZTMP3_DEVICE, ZTMP4_DEVICE )
    CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE, X_LES_RES_ddxa_W_SBG_UaW , .TRUE. )
    !
    CALL GY_M_V_DEVICE(KKA,KKU,KKL,PTHLM,PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
    CALL MZF_DEVICE( ZFLXZ, ZTMP2_DEVICE )
    !$acc kernels present_cr(ZTMP3_DEVICE)
    ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
    !$acc end kernels
    CALL MYF_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
    CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE, X_LES_RES_ddxa_Thl_SBG_UaW , .TRUE. )
    !
    IF (KRR>=1) THEN
      CALL GY_V_M_DEVICE(PRM(:,:,:,1),PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
      CALL MZF_DEVICE( ZFLXZ, ZTMP2_DEVICE )
      !$acc kernels present_cr(ZTMP3_DEVICE)
      ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
      !$acc end kernels
      CALL MYF_DEVICE(ZTMP3_DEVICE,ZTMP4_DEVICE)
      CALL LES_MEAN_SUBGRID( ZTMP4_DEVICE,X_LES_RES_ddxa_Rt_SBG_UaW , .TRUE. )
    END IF
    !
!$acc end data
#endif
    CALL SECOND_MNH(ZTIME2)
    XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
  END IF
  !
END IF
!
!
!----------------------------------------------------------------------------
!
!*       7.   DIAGNOSTIC COMPUTATION OF THE 1D <W W> VARIANCE
!             -----------------------------------------------
!
IF ( OTURB_FLX .AND. tpfile%lopened .AND. HTURBDIM == '1DIM') THEN
#ifndef MNH_OPENACC
  ZFLXZ(:,:,:)= (2./3.) * PTKEM(:,:,:)                     &
     -XCMFS*PLM(:,:,:)*SQRT(PTKEM(:,:,:))*GZ_W_M(PWM,PDZZ)
#else
  CALL GZ_W_M_DEVICE(PWM,PDZZ,ZTMP1_DEVICE)
!$acc kernels present_cr(ZFLXZ)
  ZFLXZ(:,:,:)= (2./3.) * PTKEM(:,:,:)                     &
     -XCMFS*PLM(:,:,:)*SQRT(PTKEM(:,:,:))*ZTMP1_DEVICE(:,:,:)
!$acc end kernels
#endif
  ! stores the W variance
  TZFIELD%CMNHNAME   = 'W_VVAR'
  TZFIELD%CSTDNAME   = ''
  TZFIELD%CLONGNAME  = 'W_VVAR'
  TZFIELD%CUNITS     = 'm2 s-2'
  TZFIELD%CDIR       = 'XY'
  TZFIELD%CCOMMENT   = 'X_Y_Z_W_VVAR'
  TZFIELD%NGRID      = 1
  TZFIELD%NTYPE      = TYPEREAL
  TZFIELD%NDIMS      = 3
  TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLXZ)
  CALL IO_Field_write(TPFILE,TZFIELD,ZFLXZ(:,:,:))
END IF

if ( mppdb_initialized ) then
  !Check all inout arrays
  call Mppdb_check( prus, "Turb_ver_dyn_flux end:prus" )
  call Mppdb_check( prvs, "Turb_ver_dyn_flux end:prvs" )
  call Mppdb_check( prws, "Turb_ver_dyn_flux end:prws" )
  !Check all out arrays
  call Mppdb_check( pwu, "Turb_ver_dyn_flux end:pwu" )
  call Mppdb_check( pwv, "Turb_ver_dyn_flux end:pwv" )
  call Mppdb_check( pdp, "Turb_ver_dyn_flux end:pdp" )
end if

!$acc end data

#ifndef MNH_OPENACC
deallocate (zdirsinzw,zcoefs,za,zres,zflxz,zsource,zkeff,zcoefflxu,zcoefflxv,zuslopem,zvslopem)
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'TURB_VER_DYN_FLUX' )
#endif
!$acc end data

!----------------------------------------------------------------------------
!
END SUBROUTINE TURB_VER_DYN_FLUX
